﻿Imports System.ComponentModel
Imports System.Drawing.Color
Imports System.Reflection.MethodBase
Imports System.Runtime.InteropServices
Imports ErrorLog
Imports ErrorLog.LogStates
Imports Windows.UI.ViewManagement
Imports Extensions.API.CAPI
Imports Extensions
Imports Microsoft.Win32

Public Class frmTaskbar
    Private sModule As String = "frmTaskbar"
    Public Event ErrorRaised(oDetail As ErrorDetail)

    Private bBarRegistered As Boolean = False
    Private uCallBack As Integer
    Private iDockedEdge As ABEdge
    Private iMaxWidth As Integer = 200
    Private oAccentColour As Windows.UI.Color

    Private oWindowList As New Dictionary(Of String, ucWindow)
    Private oDWM As Dwm.Thumbnail

    ' To Do
    ' Sort new window list by process ID
    ' - create dictionary by process ID with Window names, use sorted keys to copy to main window list
    ' Settings screen & control auto-run
    ' Ignore list for rogue Window names
    ' Drop target without maximising application?

    Private Sub frmTaskbar_Load(sender As Object, e As EventArgs) Handles MyBase.Load
        Try
            Me.Height = 26
            RegisterBar(ABEdge.ABE_BOTTOM)
            pnlWindows.BackColor = Color.Transparent
            frmThumbnail.Top = Me.Top - frmThumbnail.Height
            frmThumbnail.Left = 0
            'Dim sKey As String = "Software\Microsoft\Windows\CurrentVersion\Run"
            'Dim oKey As RegistryKey = My.Computer.Registry.CurrentUser.OpenSubKey(sKey)
            'oKey.SetValue("LabelBar", Application.ExecutablePath)
        Catch ex As Exception
            frmSettings.EL.LogMessage(New ErrorDetail(LSError, ex.Message, "", sModule & "." & GetCurrentMethod().Name, ex))
        End Try
    End Sub

    Private Sub frmTaskbar_Closing(sender As Object, e As MouseEventArgs) Handles cmdExit.Click
        RegisterBar(iDockedEdge)
        End
    End Sub

    Private Sub timRefresh_Tick(sender As Object, e As EventArgs) Handles timRefresh.Tick
        Try
            Dim oNewWindowList As New Dictionary(Of String, ucWindow)
            Dim sWindow As String
            Dim hWnd As Integer
            Dim oSortList As New SortedDictionary(Of String, String)

            For Each oProcess As Process In Process.GetProcesses    ' Scan running processes for those with visible windows
                sWindow = oProcess.MainWindowTitle
                If {"Taskbar Labels", "Settings", "Microsoft Text Input Application"}.Contains(sWindow) Then Continue For
                If sWindow <> "" Then
                    If oWindowList.Keys.Contains(sWindow) Then      ' We already have this window
                        oNewWindowList.Add(sWindow, oWindowList.Item(sWindow))
                        oSortList.Add(oProcess.Id.ToString("00000") & "-" & sWindow, sWindow)
                    Else                                            ' We haven't seen this one before
                        Dim oWindow As New ucWindow(oProcess)
                        AddHandler oWindow.MouseHovering, AddressOf HandleMouseHover
                        AddHandler oWindow.MouseLeft, AddressOf HandleMouseLeave
                        oNewWindowList.Add(sWindow, oWindow)
                        oSortList.Add(oProcess.Id.ToString("00000") & "-" & sWindow, sWindow)
                    End If
                End If
            Next
            For Each sWindow In oWindowList.Keys                    ' Now we need to check for windows which are still open but "main window" changed e.g. Excel
                If Not oNewWindowList.Keys.Contains(sWindow) Then   ' Not already in the new windows collection
                    hWnd = FindWindow(vbNullString, sWindow)
                    If hWnd <> 0 Then
                        oNewWindowList.Add(sWindow, oWindowList.Item(sWindow))
                        oSortList.Add(oWindowList.Item(sWindow).ProcessID.Id.ToString("00000") & "-" & sWindow, sWindow)
                    End If
                End If
            Next

            Dim bChanged As Boolean                                 ' Detect if window list has changed
            If oWindowList.Count <> oNewWindowList.Count Then       ' Definitely changed!
                bChanged = True
            Else                                                    ' Have to check keys
                For Each sKey As String In oWindowList.Keys
                    If Not oNewWindowList.Keys.Contains(sKey) Then bChanged = True
                Next
            End If
            If bChanged Then
                oWindowList = oNewWindowList                        ' Update the list of known windows
                pnlWindows.Controls.Clear()                         ' Re-populate the bar
                Dim iLeft As Integer
                Dim iWidth As Integer
                iWidth = CInt((SystemInformation.PrimaryMonitorSize.Width - cmdExit.Width) / oWindowList.Count)
                If iWidth > iMaxWidth Then iWidth = iMaxWidth
                ' For Each oControl As ucWindow In oWindowList.Values
                For Each sWindow In oSortList.Values
                    Dim oControl As ucWindow = oNewWindowList(sWindow)
                    pnlWindows.Controls.Add(oControl)
                    oControl.Left = iLeft
                    oControl.Width = iWidth - 3                    ' Space between items
                    iLeft += iWidth
                Next
            End If
            CheckColours()                                          ' Check colour scheme & adjust
        Catch ex As Exception
            frmSettings.EL.LogMessage(New ErrorDetail(LSError, ex.Message, "", sModule & "." & GetCurrentMethod().Name, ex))
        End Try
    End Sub

    Private Sub HandleMouseHover(oWindow As ucWindow)
        ' Pop up thumbnail and full text
        Try
            frmThumbnail.lblText.Text = oWindow.Title
            ' Debug.Print(oWindow.Title)
            frmThumbnail.Top = Me.Top - frmThumbnail.Height
            frmThumbnail.Left = oWindow.Left
            frmThumbnail.Show()
            RegisterThumbnail(oWindow.HWnd)
        Catch ex As Exception
            frmSettings.EL.LogMessage(New ErrorDetail(LSError, ex.Message, "", sModule & "." & GetCurrentMethod().Name, ex))
        End Try
    End Sub

    Private Sub HandleMouseLeave(oWindow As ucWindow)
        ' Hide Thumbnail
        Try
            frmThumbnail.Hide()
            'oDWM.SetVisible(False)
        Catch ex As Exception
            frmSettings.EL.LogMessage(New ErrorDetail(LSError, ex.Message, "", sModule & "." & GetCurrentMethod().Name, ex))
        End Try
    End Sub

    Private Sub RegisterThumbnail(hWnd As Integer)
        ' Sets up a relationship between source DWM window and our target display, then updates
        ' https://learn.microsoft.com/en-us/windows/win32/dwm/thumbnail-ovw
        ' https://github.com/vladris/DWM
        Try
            oDWM = New Dwm.Thumbnail(frmThumbnail, hWnd)
            Dim oLocation As Point = frmThumbnail.pnlInner.Location
            oLocation.Offset(frmThumbnail.pnlBorder.Padding.All, frmThumbnail.pnlBorder.Padding.All)
            oDWM.SetDestinationRectangle(New Rectangle(oLocation, frmThumbnail.picThumbnail.Size))
            oDWM.SetVisible(True)
        Catch ex As Exception
            frmSettings.EL.LogMessage(New ErrorDetail(LSError, ex.Message, "", sModule & "." & GetCurrentMethod().Name, ex))
        End Try
    End Sub

    Private Function CheckColours()
        ' Check to see if accent colour has changed, if so adjust our colours
        Try
            Dim oUISettings As New UISettings                       ' Get and copy accent colour
            Dim oA As Windows.UI.Color = oUISettings.GetColorValue(UIColorType.Accent)
            If oA <> oAccentColour Then                        ' Colour scheme has changed
                oAccentColour = oA
                Dim oAccent As Color = Color.FromArgb(oA.A, oA.R, oA.G, oA.B)
                Dim iMax As Integer = oA.R
                Dim oDark As Color = Darken(oAccent, 100)
                Me.BackColor = oDark
                frmThumbnail.pnlBorder.BackColor = oDark
                cmdExit.BackColor = oDark
                For Each oLabel As ucWindow In oWindowList.Values
                    oLabel.pnlBottom.BackColor = oAccent
                Next
            End If
        Catch ex As Exception
            frmSettings.EL.LogMessage(New ErrorDetail(LSError, ex.Message, "", sModule & "." & GetCurrentMethod().Name, ex))
        End Try
    End Function

    Private Function Darken(oColor As Color, iMax As Integer) As Color
        Try
            Dim iRange As Integer = oColor.R
            If oColor.G > iMax Then iMax = oColor.G
            If oColor.B > iMax Then iMax = oColor.B
            If iRange > iMax Then
                Dim dFactor As Double = iMax / iRange
                oColor = Color.FromArgb(oColor.A, oColor.R * dFactor, oColor.G * dFactor, oColor.B * dFactor)
            End If
            Return oColor
        Catch ex As Exception
            frmSettings.EL.LogMessage(New ErrorDetail(LSError, ex.Message, "", sModule & "." & GetCurrentMethod().Name, ex))
        End Try
    End Function

    Private Sub RegisterBar(iEdge As ABEdge)
        ' Register as an "app bar" and reserve space at bottom of screen
        Try
            Dim abd As New APPBARDATA()
            abd.cbSize = Marshal.SizeOf(abd)
            abd.hWnd = Me.Handle
            If Not bBarRegistered Then
                uCallBack = RegisterWindowMessage("AppBarMessage")
                abd.uCallbackMessage = uCallBack
                Dim ret As System.UInt32 = SHAppBarMessage(CInt(ABMsg.ABM_NEW), abd) 'ToDo: Unsigned Integers not supported
                bBarRegistered = True
                ABSetPos(iEdge)
                iDockedEdge = iEdge
            Else
                SHAppBarMessage(CInt(ABMsg.ABM_REMOVE), abd)
                bBarRegistered = False
            End If
        Catch ex As Exception
            frmSettings.EL.LogMessage(New ErrorDetail(LSError, ex.Message, "", sModule & "." & GetCurrentMethod().Name, ex))
        End Try
    End Sub 'RegisterBar

    Private Sub ABSetPos(iEdge As ABEdge)
        ' set position of appbar
        Try
            Dim abd As New APPBARDATA()
            abd.cbSize = Marshal.SizeOf(abd)
            abd.hWnd = Me.Handle
            abd.uEdge = iEdge

            If abd.uEdge = CInt(ABEdge.ABE_LEFT) Or abd.uEdge = CInt(ABEdge.ABE_RIGHT) Then
                abd.rc.top = 0
                abd.rc.bottom = SystemInformation.PrimaryMonitorSize.Height
                If abd.uEdge = CInt(ABEdge.ABE_LEFT) Then
                    abd.rc.left = 0
                    abd.rc.right = Size.Width
                Else
                    abd.rc.right = SystemInformation.PrimaryMonitorSize.Width
                    abd.rc.left = abd.rc.right - Size.Width
                End If

            Else
                abd.rc.left = 0
                abd.rc.right = SystemInformation.PrimaryMonitorSize.Width
                If abd.uEdge = CInt(ABEdge.ABE_TOP) Then
                    abd.rc.top = 0
                    abd.rc.bottom = Size.Height
                Else
                    abd.rc.bottom = SystemInformation.PrimaryMonitorSize.Height
                    abd.rc.top = abd.rc.bottom - Size.Height
                End If
            End If

            ' Query the system for an approved size and position.
            SHAppBarMessage(CInt(ABMsg.ABM_QUERYPOS), abd)

            ' Adjust the rectangle, depending on the edge to which the
            ' appbar is anchored.
            Select Case abd.uEdge
                Case CInt(ABEdge.ABE_LEFT)
                    abd.rc.right = abd.rc.left + Size.Width
                Case CInt(ABEdge.ABE_RIGHT)
                    abd.rc.left = abd.rc.right - Size.Width
                Case CInt(ABEdge.ABE_TOP)
                    abd.rc.bottom = abd.rc.top + Size.Height
                Case CInt(ABEdge.ABE_BOTTOM)
                    abd.rc.top = abd.rc.bottom - Size.Height
            End Select

            ' Pass the final bounding rectangle to the system.
            SHAppBarMessage(CInt(ABMsg.ABM_SETPOS), abd)

            ' Move and size the appbar so that it conforms to the
            ' bounding rectangle passed to the system.
            MoveWindow(abd.hWnd, abd.rc.left, abd.rc.top, abd.rc.right - abd.rc.left, abd.rc.bottom - abd.rc.top, True)
        Catch ex As Exception
            frmSettings.EL.LogMessage(New ErrorDetail(LSError, ex.Message, "", sModule & "." & GetCurrentMethod().Name, ex))
        End Try
    End Sub 'ABSetPos

    Protected Overrides Sub WndProc(ByRef m As System.Windows.Forms.Message)
        If m.Msg = uCallBack Then
            Select Case m.WParam.ToInt32()
                Case CInt(ABNotify.ABN_POSCHANGED)
                    ABSetPos(iDockedEdge)
            End Select
        End If

        MyBase.WndProc(m)
    End Sub 'WndProc

    Protected Overrides ReadOnly Property CreateParams() As System.Windows.Forms.CreateParams
        Get
            Dim cp As CreateParams = MyBase.CreateParams
            cp.Style = cp.Style And Not &HC00000 ' WS_CAPTION
            cp.Style = cp.Style And Not &H800000 ' WS_BORDER
            cp.ExStyle = &H80 Or &H8 ' WS_EX_TOOLWINDOW | WS_EX_TOPMOST
            Return cp
        End Get
    End Property

End Class
